#include "precompile.h"

#include <openssl/bn.h>

#include <google/protobuf/message.h>
#include <a8/mixedsession.h>
#include <a8/tcplistener.h>

#include "framework/cpp/netmsghandler.h"

#include "app.h"
#include "GCListener.h"
#include "jsondatamgr.h"
#include "handlermgr.h"
#include "cs_proto.pb.h"
#include "cs_msgid.pb.h"

const static size_t CRYPTED_SEND_LEN = 4;
const static size_t CRYPTED_RECV_LEN = 6;

class GCClientSession: public a8::MixedSession
{
public:

    virtual void Reset() override
    {
        a8::MixedSession::Reset();
        seed_ = rand();
        send_i_ = 0;
        send_j_ = 0;
        recv_i_ = 0;
        recv_j_ = 0;
        initialized_ = false;
    }

    void DecryptRecv(unsigned char* data, size_t len)
    {
        if (!initialized_) {
            return;
        }
        if (len < CRYPTED_RECV_LEN) {
            return;
        }
        for (size_t t = 0; t < CRYPTED_RECV_LEN; t++) {
            recv_i_ %= key_.size();
            unsigned char x = (data[t] - recv_j_) ^ key_[recv_i_];
            ++recv_i_;
            recv_j_ = data[t];
            data[t] = x;
        }
    }

    void EncryptSend(unsigned char* data, size_t len)
    {
        if (!initialized_) {
            return;
        }

        if (len < CRYPTED_SEND_LEN) {
            return;
        }

        for (size_t t = 0; t < CRYPTED_SEND_LEN; t++) {
            send_i_ %= key_.size();
            unsigned char x = (data[t] ^ key_[send_i_]) + send_j_;
            ++send_i_;
            data[t] = send_j_ = x;
        }
    }

    virtual void DecodeUserPacket(char* buf, int& offset, unsigned int buflen) override
    {
        #if 1
        is_activite = true;
        #endif
        bool warning = false;
        while (buflen - offset >= sizeof(ClientPktHeader)) {
            ClientPktHeader header = *((ClientPktHeader*)&buf[offset]);
            DecryptRecv((unsigned char*)&header, sizeof(header));
            header.size =
                ((header.size & 0x00FF) << 8) +
                ((header.size & 0xFF00) >> 8);
            if (buflen - offset < sizeof(ClientPktHeader) + header.size - 4) {
                break;
            }
            header.size -= 4;
            App::Instance()->AddSocketMsg(SF_GameClient,
                                          socket_handle,
                                          saddr,
                                          header.cmd,
                                          0,
                                          &buf[offset + sizeof(ClientPktHeader)],
                                          header.size);
            offset += sizeof(ClientPktHeader) + header.size;
        }

        if (warning) {
            a8::UdpLog::Instance()->Warning("收到client非法数据包", {});
        }
    }

    virtual void OnRawHttpGet(const std::string& url, const std::string& querystr,
                              std::string& response) override
    {
        App::Instance()->AddIMMsg(IM_ExecGM,
                                  a8::XParams()
                                  .SetSender(socket_handle)
                                  .SetParam1(url)
                                  .SetParam2(querystr)
                                  .SetParam3(saddr)
                                  );
    }

    virtual void OnConnect() override
    {
        cs::SMAuthChallenge msg;
        msg.set_seed(seed_);
        unsigned short packlen = 2 + CustomPbSerializeSize(&msg);

        char* buff = (char*)malloc(sizeof(ServerPktHeader) + packlen);
        ServerPktHeader* head = (ServerPktHeader*)buff;
        head->cmd = cs::_SMAuthChallenge;
        head->size = (packlen << 8) + (packlen >> 8);
        int offset = 0;
        CustomPbSerialize(buff + sizeof(ServerPktHeader), offset, packlen, &msg);
        SendBuff(buff, sizeof(ServerPktHeader) + packlen - 2);
        free(buff);
    }

    virtual void OnDisConnect() override
    {
        App::Instance()->AddIMMsg(IM_ClientSocketDisconnect,
                                  a8::XParams()
                                  .SetSender(socket_handle)
                                  .SetParam1(1));
    }

    void SetKey(unsigned char* data, int len)
    {
        send_i_ = 0;
        send_j_ = 0;
        recv_i_ = 0;
        recv_j_ = 0;
        initialized_ = true;
        key_.clear();
        for (int i = 0; i < len; ++i) {
            key_.push_back(data[i]);
        }
    }

protected:

    virtual void SendBuff(const char* buff, unsigned int bufflen) override
    {
        EncryptSend((unsigned char*)buff, bufflen);
        a8::MixedSession::SendBuff(buff, bufflen);
    }

private:
    unsigned int seed_ = 0;
    std::vector<unsigned char> key_;
    unsigned char send_i_ = 0;
    unsigned char send_j_ = 0;
    unsigned char recv_i_ = 0;
    unsigned char recv_j_ = 0;
    bool initialized_ = false;
};

static void CreateGameClientSocket(a8::TcpSession **p)
{
    *p = new GCClientSession();
}

static void GSListeneron_error(a8::TcpListener*, int type, int errorid)
{
    abort();
    a8::UdpLog::Instance()->Debug("GCListeneron_error %d %d", {type, errorid});
}

void GCListener::Init()
{
    tcp_listener_ = new a8::TcpListener();
    tcp_listener_->on_create_client_socket = CreateGameClientSocket;
    tcp_listener_->on_error = GSListeneron_error;

    tcp_listener_->bind_address = "0.0.0.0";
    #if 1
    tcp_listener_->bind_port = 8086;
    #else
    tcp_listener_->bind_port = JsonDataMgr::Instance()->GetConf()->At("listen_port")->AsXValue();
    #endif
    tcp_listener_->Open();
}

void GCListener::UnInit()
{
    delete tcp_listener_;
    tcp_listener_ = nullptr;
}

void GCListener::SendText(int sockhandle, const std::string& text)
{
    tcp_listener_->SendClientMsg(sockhandle, text.data(), text.size());
}

void GCListener::ForceCloseClient(int sockhandle)
{
    tcp_listener_->ForceCloseClient(sockhandle);
}

void GCListener::MarkClient(int sockhandle, bool is_active)
{
    tcp_listener_->MarkClient(sockhandle, is_active);
}

void GCListener::SetKey(int sockhandle, const std::string& key)
{
    tcp_listener_->LockClients();
    GCClientSession* session = (GCClientSession*)tcp_listener_->GetClientSession(sockhandle);
    if (session) {
        bignum_st* bn = BN_new();
        BN_hex2bn(&bn, key.c_str());
        {
             int length = BN_num_bytes(bn);
             unsigned char* _array = new unsigned char[length];
             BN_bn2bin(bn, (unsigned char*)_array);
             std::reverse(_array, _array + length);
             session->SetKey(_array, 40);
        }
        BN_free(bn);
    }
    tcp_listener_->UnLockClients();
}

void GCListener::InternalSendMsg(int socket_handle, int msgid, google::protobuf::Message& msg)
{
#if 0
    if (msgid == cs::_SMCharEnum) {
        char bin_buf[1024] = {0};
        FILE* fp = fopen("/root/1.bin", "rb");
        fread(bin_buf, 1, 168, fp);
        fclose(fp);

        unsigned short packlen = 2 + 168;
        char* buff = (char*)malloc(sizeof(ServerPktHeader) + packlen);
        memset(buff, 0, sizeof(ServerPktHeader) + packlen);
        ServerPktHeader* head = (ServerPktHeader*)buff;
        head->cmd = msgid;
        head->size = (packlen << 8) + (packlen >> 8);
        memmove(buff + sizeof(ServerPktHeader), bin_buf, packlen - 2);
        tcp_listener_->SendClientMsg(socket_handle, buff, sizeof(ServerPktHeader) + packlen);
        free(buff);
    } else {
#else
        {
#endif
        unsigned short packlen = 2 + CustomPbSerializeSize(&msg);
        char* buff = (char*)malloc(sizeof(ServerPktHeader) + packlen);
        memset(buff, 0, sizeof(ServerPktHeader) + packlen);
        ServerPktHeader* head = (ServerPktHeader*)buff;
        head->cmd = msgid;
        head->size = (packlen << 8) + (packlen >> 8);
        int offset = 0;
        CustomPbSerialize(buff + sizeof(ServerPktHeader), offset, packlen, &msg);
        tcp_listener_->SendClientMsg(socket_handle, buff, sizeof(ServerPktHeader) + packlen - 2);
        free(buff);
    }
}
